API_KEY_ENV: str = "CONFIDENT_AI_API_KEY"
LOG_TO_SERVER_ENV: str = "DO_NOT_SEND_TO_CONFIDENT_AI"
KEY_FILE: str = ".deepeval"
PYTEST_TEMP_FILE_NAME_ENV_VAR: str = "CONFIDENT_AI_RUN_TIMESTAMP"
PYTEST_RUN_TEST_NAME: str = "CONFIDENT_AI_RUN_TEST_NAME"
